% ===   ***** ***** ***** ***** ***** ***** *****  	=== %
% ===   BSPTREE_INITILIZATION                       === %
% ===   ***** ***** ***** ***** ***** ***** *****  	=== %
% === 	CREATED: 22 NOVEMBER, 2014                 	=== %
% ===   UPDATED: 06 AUGUST, 2015                    === %
% ===   ***** ***** ***** ***** ***** ***** *****  	=== %

function [bsptree, cur_tree_size] = bsptree_initilization(dim, max_tree_size, special)
% === === === === === === === === === %
% INPUT:  1) DIMENSION;
%         2) THE NUMBER OF TREE NODES IN TOTAL;
%         3) SPECIAL TYPE OF BSPTREE, FOR NON-PRUNE, LUR-PRUNE, OR RND-PRUNE;
% OUTPUT: 1) THE BSPTREE PER SE;
%         2) THE CURRENT TREE NODE NUMBER (SIZE), WHICH IS 0 AFTER INITIALIZATION;
% === === === === === === === === === %

switch special
    case 'lru'
        bsptree = struct ('x',      zeros(dim,  max_tree_size),     ...     % SOLUTION %
                          'f',      zeros(1,    max_tree_size),     ...     % FITNESS OF SOLUTION %
                          'd',      -1 * ones(1,max_tree_size),     ...     % PARTITIONING ON THIS DIMENSION %
                          'min',    zeros(1,    max_tree_size),   	...     % LOWER BOUND OF SUB-REGION %
                          'max',    zeros(1,    max_tree_size),   	...     % UPPER BOUND OF SUB-REGION %
                          'p_idx',  zeros(1,    max_tree_size),     ...     % POINTER TO PARENT %
                          'l_idx',  (max_tree_size + 1) * ones(1, max_tree_size),	...     % POINTER TO LEFT CHILD %
                          'r_idx',  (max_tree_size + 1) * ones(1, max_tree_size),	...     % POINTER TO RIGHT CHILD %
                          'ts',     zeros(1,    max_tree_size),     ...     % TIME STAMP FOR LRU PRUNE %
                          'last',   max_tree_size + 1,              ...     % A POITER THAT POINTS TO NULL %
                          'vn',     zeros(1,    max_tree_size)      ...     % IS A VIRTUAL NODE OR NOT %
                          );
                      bsptree.ts(1, bsptree.last) = inf;
    case 'rnd'  % RAND WITHOUT .TS, BUT TS FOR TRAVESAL OF TREE TO GET A RAND TO-PRUNE-LIST %
        bsptree = struct ('x',      zeros(dim,  max_tree_size),     ...
                          'f',      zeros(1,    max_tree_size),     ...
                          'd',      -1 * ones(1,max_tree_size),     ...
                          'min',    zeros(1,    max_tree_size),   	...
                          'max',    zeros(1,    max_tree_size),   	...
                          'p_idx',  zeros(1,    max_tree_size),     ...
                          'l_idx',  (max_tree_size + 1) * ones(1, max_tree_size),	...
                          'r_idx',  (max_tree_size + 1) * ones(1, max_tree_size),	...
                          'ts',     zeros(1,    max_tree_size),     ...
                          'last',   max_tree_size + 1,              ...
                          'vn',     zeros(1,    max_tree_size)      ...
                          );
                      bsptree.ts(1, bsptree.last) = inf;
    case 'non'  % FOR ORIGINAL CNRGA BSPTREE, NO .TS, .LAST, .VN %
        bsptree = struct ('x',      zeros(dim,  max_tree_size),     ...
                          'f',      zeros(1,    max_tree_size),     ...
                          'd',      -1 * ones(1,max_tree_size),     ...
                          'min',    zeros(1,    max_tree_size),   	...
                          'max',    zeros(1,    max_tree_size),   	...
                          'p_idx',  zeros(1,    max_tree_size),     ...
                          'l_idx',  (max_tree_size + 1) * ones(1, max_tree_size),	...
                          'r_idx',  (max_tree_size + 1) * ones(1, max_tree_size)	...
                          );  %NO .TS, .LAST, .VN %
end
    bsptree.d(1, 1) = -2;
    cur_tree_size = 0;
    % === RULES === %
    % 1) ALL UN-INITIALIZED NODES: [.d == -1]
    % 2) ALL LEAF NODES: 2.1) [.d == 0]
end

