% ===   ***** ***** ***** ***** ***** ***** *****  	=== %
% ===   BSPTREE_INSERTION RAND PRUNING              === %
% ===   ***** ***** ***** ***** ***** ***** *****  	=== %
% ===   UPDATED: 01 NOVEMBER, 2015                  === %
% ===   ***** ***** ***** ***** ***** ***** *****  	=== %

function [cur_bsptree_node, insert_pos] = bsptree_insertion_r(x, f, cur_bsptree_node, target_pos)

global BSPTREE;

[~, pop_size] = size(x);
insert_pos = zeros(1, pop_size);

for idx = 1 : pop_size
    if BSPTREE.d(1, 1) == -2  % ROOT NODE ONLY %
        BSPTREE.x(:, 1)  = x(:, 1);
        BSPTREE.f(1, 1)  = f(1, 1);
        BSPTREE.d(1, 1)  = 0;
        BSPTREE.vn(1, 1) = 1;
        cur_bsptree_node = 1;
        insert_pos(1, 1) = 1;
    else
        n_idx = 1;
        while BSPTREE.d(1, n_idx)
            jdx = BSPTREE.d(1, n_idx);
            if x(jdx, idx) < (BSPTREE.min(1, n_idx) + BSPTREE.max(1, n_idx)) / 2.0
                n_idx = BSPTREE.l_idx(1, n_idx);
            else
                n_idx = BSPTREE.r_idx(1, n_idx);
            end
        end  % THEN PARTITION THE NODE 'NIDX' %
        
        insert_pos(1, idx) = n_idx;
        memo_x = BSPTREE.x(:, n_idx);
        memo_f = BSPTREE.f(1, n_idx);
        
        d_diff = abs(memo_x - x(:, idx));
        d = find(d_diff == max(d_diff));
        d = d(1);
% %         if length(d) > 1
% %             d = d(ceil(rand * length(d)));
% %         end

        cur_min = min(memo_x(d, 1), x(d, idx));
        cur_max = max(memo_x(d, 1), x(d, idx));
        
        BSPTREE.d(1, n_idx) = d;
        BSPTREE.min(1, n_idx) = cur_min;
        BSPTREE.max(1, n_idx) = cur_max;
        
        if nargin <= 3  % NO 'TARGET_POS' %
            l_idx = cur_bsptree_node + 1;
            r_idx = cur_bsptree_node + 2;
        else
            l_idx = target_pos(2 * idx - 1);
            r_idx = target_pos(2 * idx);
        end
        
        BSPTREE.l_idx(1, n_idx) = l_idx;
        BSPTREE.r_idx(1, n_idx) = r_idx;
        BSPTREE.p_idx(1, l_idx) = n_idx;
        BSPTREE.p_idx(1, r_idx) = n_idx;
        BSPTREE.d(1, l_idx) = 0;
        BSPTREE.d(1, r_idx) = 0;
        
        if nargin <= 3
            cur_bsptree_node = cur_bsptree_node + 2;
        end
        
        if x(d, idx) < (cur_min + cur_max) / 2.0
            BSPTREE.x(:, l_idx) = x(:, idx);
            BSPTREE.f(1, l_idx) = f(1, idx);
            BSPTREE.x(:, r_idx) = memo_x;
            BSPTREE.f(1, r_idx) = memo_f;
            BSPTREE.vn(1, r_idx) = 1;
            BSPTREE.vn(1, l_idx) = 0;
        else
            BSPTREE.x(:, r_idx) = x(:, idx);
            BSPTREE.f(1, r_idx) = f(1, idx);
            BSPTREE.x(:, l_idx) = memo_x;
            BSPTREE.f(1, l_idx) = memo_f;
            BSPTREE.vn(1, l_idx) = 1;
            BSPTREE.vn(1, r_idx) = 0;
        end
    end  % END OF IF-ELSE 'BSPTREE.D(1, 1) == -2' %

end  % END OF 'FOR IDX = 1 : POP_SIZE' %

end